#define _USE_MATH_DEFINES
#include <cmath>
#include <iostream>
#include <filesystem>
#include <set>

#define TINYOBJLOADER_IMPLEMENTATION
#include "tiny_obj_loader.h"

using namespace std;
namespace fs = std::filesystem;
double tol_equal = 1e-8;

double round_to(double value, double precision = 1.0)
{
	// https://stackoverflow.com/questions/14369673/round-double-to-3-points-decimal
    return std::round(value / precision) * precision;
}

struct Vertex
{
	double x;
	double y;
	double z;

	Vertex(double _x, double _y, double _z)
		: x(_x), y(_y), z(_z)
	{
		x = round_to(x, 1e-4);
		y = round_to(y, 1e-4);
		z = round_to(z, 1e-4);
	}
	
	void print(ostream &out = cout) const
	{
		out << x << " " << y << " " << z;
	}
	
	int hash(int n) const
	{
		// Optimized Spatial Hashing for Collision Detection of Deformable Objects
		const int p1 = 73856093;
		const int p2 = 19349663;
		const int p3 = 83492791;
		return ((int)(x*p1) ^ (int)(y*p2) ^ (int)(z*p3)) % n;
	}
};

ostream & operator<<(ostream &out, const Vertex &rhs)
{
	rhs.print(out);
	return out;
}

inline bool operator==(const Vertex& left, const Vertex& right)
{
	double dx = left.x - right.x;
	double dy = left.y - right.y;
	double dz = left.z - right.z;
	double norm = sqrt(dx*dx + dy*dy + dz*dz);
	if(norm > tol_equal)
		return false;
	else
		return true;
}

inline bool operator<(const Vertex& left, const Vertex& right)
{
	if (abs(left.x - right.x) < tol_equal) {
		if(abs(left.y - right.y) < tol_equal) {
			if(abs(left.z - right.z) < tol_equal) {
				//Strict weak ordering
				return false;
			}
			else if (left.z < right.z)
				return true;
			else
				return false;
		}
		else if (left.y < right.y)
			return true;
		else
			return false;
	}
	else if (left.x < right.x)
		return true;
	else
		return false;	
}

bool equalsTol(const Vertex &left, const Vertex &right, double tol)
{
	double dx = left.x - right.x;
	double dy = left.y - right.y;
	double dz = left.z - right.z;
	double norm = sqrt(dx*dx + dy*dy + dz*dz);
	return norm < tol;
}

void loadOBJ(const string &meshName, vector<float> &posBuf)
{
	// Load geometry
	tinyobj::attrib_t attrib;
	std::vector<tinyobj::shape_t> shapes;
	std::vector<tinyobj::material_t> materials;
	string warnStr, errStr;
	bool rc = tinyobj::LoadObj(&attrib, &shapes, &materials, &warnStr, &errStr, meshName.c_str());
	if(!rc) {
		cerr << errStr << endl;
	} else {
		posBuf = attrib.GetVertices();
	}
}

bool objscmp(const string &folder1, const string &folder2, double tol)
{
	// https://stackoverflow.com/questions/62409409/how-to-make-stdfilesystemdirectory-iterator-to-list-filenames-in-order
	set<fs::path> sortedFiles1;
	set<fs::path> sortedFiles2;
	cout << folder1 << endl;
	for(const auto &entry : fs::directory_iterator(folder1)) {
		if(entry.path().extension() == ".obj") {
			//cout << entry.path() << endl;
			sortedFiles1.insert(entry.path());
		}
	}
	cout << folder2 << endl;
	for(const auto &entry : fs::directory_iterator(folder2)) {
		if(entry.path().extension() == ".obj") {
			//cout << entry.path() << endl;
			sortedFiles2.insert(entry.path());
		}
	}
	if(sortedFiles1.size() != sortedFiles2.size()) {
		cout << "The directories do not have the same number of obj files." << endl;
		return false;
	}
	// Now the directories do have the same number of files
	
	// Run 2 tests:
	// Test0: Just compare the files line-by-line
	// Test1: Do an n^2 with hashing
	bool pass0 = false;
	bool pass1 = false;
	for(int test = 0; test < 2; ++test) {
		int nobjs = (int)sortedFiles1.size();
		auto file1 = sortedFiles1.cbegin();
		auto file2 = sortedFiles2.cbegin();
		int numNotMatchSum = 0;
		for(int k = 0; k < nobjs; ++k) {
			cout << file1->filename() << ": ";
			if(file1->filename() != file2->filename()) {
				cout << "Filenames do not match: " << file1->filename() << ", " << file2->filename() << endl;
			}
			//cout << entry1->filename() << endl;
			vector<float> posBuf1;
			vector<float> posBuf2;
			loadOBJ(file1->string(), posBuf1);
			loadOBJ(file2->string(), posBuf2);
			if(posBuf1.size() != posBuf2.size()) {
				cout << file1->filename() << " and " << file2->filename() << " do not have the same number of vertices." << endl;
				return false;
			}
			int nverts = (int)posBuf1.size()/3;
			vector<Vertex> vertBuf1;
			vector<Vertex> vertBuf2;

			for(int i = 0; i<nverts; ++i) {
				vertBuf1.push_back(Vertex(posBuf1[3*i + 0], posBuf1[3*i + 1], posBuf1[3*i + 2]));
				vertBuf2.push_back(Vertex(posBuf2[3*i + 0], posBuf2[3*i + 1], posBuf2[3*i + 2]));
			}
			
			bool match = true;
			if(test == 0) {
				// Linear matching
				for(int i = 0; i < vertBuf1.size(); i++) {
					if(!equalsTol(vertBuf1[i], vertBuf2[i], tol)) {
						match = false;
						break;
					}
				}
			} else {
				// Quadratic matching
				for(auto v1 : vertBuf1) {
					for(auto v2 : vertBuf2) {
						if(equalsTol(v1, v2, tol)) {
							match = true;
							break;
						}
					}
					if(match) {
						break;
					}
				}
				if(match) {
					for(auto v2 : vertBuf2) {
						for(auto v1 : vertBuf1) {
							if(equalsTol(v2, v1, tol)) {
								match = true;
								break;
							}
						}
						if(match) {
							break;
						}
					}
				}
			}
			
			if(test == 0) {
				cout << "   linear";
				if(match) {
					cout << " PASS";
				} else {
					cout << " ...";
					++numNotMatchSum;
				}
			} else {
				cout << "quadratic";
				if(match) {
					cout << " PASS";
				} else {
					cout << " FAIL";
					++numNotMatchSum;
				}
			}
			
			cout << endl;
			++file1;
			++file2;
		}
		
		if(test == 0) {
			pass0 = (numNotMatchSum == 0);
			if(pass0) {
				// If Test0 passed, then we pass the test, so break
				break;
			}
		} else {
			pass1 = (numNotMatchSum == 0);
		}
	}
	
	return pass0 || pass1;
}

int main(int argc, char **argv)
{
	if(argc < 3) {
		cout << "Usage: objscmp <FOLDER1> <FOLDER2> <OPTIONAL TOL>" << endl;
		return 0;
	}

	string folder1(argv[1]);
	string folder2(argv[2]);
	double tol = 1e-3;
	if(argc >= 4) {
		tol = atof(argv[3]);
	}

	bool pass = objscmp(folder1, folder2, tol);
	cout << (pass ? "PASS" : "FAIL") << endl;

	return 0;
}
